#!/usr/bin/perl
#
# Top-level perl script for CSPP_MIRS package.
# Uses only Perl modules that are part of the v5.006 standard install.
#
# A set of user-defined functions are appended to this main script;
# they are listed in alphabetical order.
#
# Pre-requisites:
# MIRS distribution in directory $CSPP_MIRS_HOME/$CSPP_MIRS_REVNAME
# Static-linked binaries in the $CSPP_MIRS_HOME/$CSPP_MIRS_REVNAME/bin
# directory, various PCFs (paths & configuration files) in the 
# $CSPP_MIRS_HOME/$CSPP_MIRS_REVNAME/setup directory to serve
# as templates, various SCSs (sequence control scripts) also to serve as 
# templates in the $CSPP_MIRS_REVNAME/scripts directory. Template files are
# modified to become runtime scripts are prepended 'cspp_', for example
# setup/npp_pcf.bash is used as a template to create cspp_npp_pcf.bash;
# scripts/npp_scs.bash is a template used to create cspp_npp_scs.bash.
# These two bash scripts are then used to control the end-to-end processing
# through a number of steps in the MIRS processing chain, namely ...
#
#   step_rdr2tdr                 # RDR->TDR
#   step_mergeNedt               # MERGE NEDTs
#   step_tdr2sdr                 # TDR->SDR
#   step_fm                      # FOOTPRINT MATCHING
#   step_externalDataFromRegress # USE REGRESSION ALGS TO GENERATE EXTERN DATA
#   step_fmsdr2edr               # FMSDR->EDR
#   step_prepSweClimo            # PREPARE AND GENERATE THE COLLOCATED SWE
#                                  CLIMATOLOGY FILES FOR VIPP (new for 2.0)
#   step_vipp                    # VERTICAL INTEGRATION AND POST PROCESSING
#   step_sfr                     # SNOWFALL RATE PRODUCT (optional, default)
#   step_nc                      # Convert EDR & DEP into NETCDF4 format
#   step_clean                   # DISK CLEANING/PURGING
#
# ... which is the subset of MIRS processing steps used in operations.
# For step_clean I actually set it to zero and implement my own cleaning
# procedure; just remove the temporary working directory at end of job.
#
# Look for the !!!!! to find areas that need attention !!!!!
#
#
# START-HELP (N.B. This tag is used in printing help, do not touch!)
#
#  Options:
#  --version             show program's version number and exit
#  ?, -h, --help         show this help message and exit
#  -n, --nothing         make the PCF and SCS files, but do not execute.
#
#  Mandatory Arguments:
#    At a minimum these arguments must be specified [i.e. no default].
#
#    -i INPUT_FILES, --input_files=INPUT_FILES
#                        Fully qualified path to input files. May be directory
#                        name, or directory/filemask (use * and ?, not regexp).
#                        Filemasks MUST be "quoted" so OS does NOT expand!
#                        e.g. -i "/dataDir/{GATMO,SATMS,TATMS}_npp*t1823486*.h5"
#                        OR -i "/dataDir/{amsua,mhs}l1b_noaa18_*_1051*l1b" OR
#                        -i "/dataDir/*.{AMA,MHS}X.M2.D13329.S0346.E0529.B*.SV"
#                        (BUT -i /dataDir  really is the simplest way).
#    -s MISSION, --sat=MISSION
#                        The mission satellite data for MIRS algorithm.
#                        Possible values (following MIRS documentation) ...
#                        'npp', 'n18', 'n19', 'n20', 'metopA', 'metopB'.
#
#  Extra Options:
#    These options may be used to customize behaviour of this program.
#
#    -w WORK_DIR, --work=WORK_DIR
#                        The directory in which all activity will occur,
#                        this is where output files will be put.
#                        [default: current directory].
#    -d DYNAMIC_ANC, --dynanc=DYNAMIC_ANC
#                        Path to root of dynamic ancillary data tree;
#                        beneath here ancillary data are organized as
#                        per the SSEC remote ancillary server for CSPP,
#                        https://jpssdb.ssec.wisc.edu/cspp_v_2_0/ancillary/,
#                        i.e. %Y_%m_%d_%j where %Y is four-digit year, %m is
#                        two-digit month, %d is two-digit day of month and
#                        %j is three-digit day of year.
#                        [CSPP_MIRS_HOME/data/dynamic_ancillary]
#    -r RES, --resolution=RES
#                        Executes retrieval in RES resolution mode for sensors
#                        aboard 'n18', 'n19', 'metopA', and 'metopB'.  RES is
#                        'HI' or 'LO'.  When 'LO' the higher resolution sensor
#                        (MHS) is aggregated to the lower resolution sensor
#                        (AMSUA) footprint. 'HI' means AMSUA interpolated to
#                        MHS footprint.
#                        [default: 'HI'].
#    -p PROCESSORS, --processors=PROCESSORS
#                        Number of cpus to use for MIRS processing.
#                        [default: --processors=1, max is 32].
#    -b BLOCK, --block=BLOCK
#                        Choose bias correction file for block BLOCK SDRs for
#                        S-NPP and NOAA-20 ATMS only.  Represents ADL Block 1
#                        or Block 2 software version.  ADL block 2 SDRs in
#                        effect from about March 2017.
#                        [default: 2; the only other valid option is 1].
#    -f, --SFR           Generate SnowFall Rate (SFR).  The only step that
#                        requires ancillary (GFS) data to be fetched, so we
#                        recommended --noSFR if you do not need this product.
#                        [default: --SFR, --noSFR is the contrary option].
#    -x, --xfiles        Use the NOAA filenames for AMSUA/MHS L1B files
#                        [default: Direct Broadcast filenames].
#    -v, --verbose       Each occurrence increases verbosity 1 level from
#                        [default: ERROR only] -v=WARNING -vv=INFO -vvv=DEBUG
#    -c, --clean         Deletes temporary working directory under WORK_DIR
#                        at process completion.
#                        [default: --clean; --noclean is the contrary option].
#    -q, --quarantine    Processing machine is not connected to the web for the
#                        purpose of fetching required dynamic ancillary data.
#                        [default: --noquarantine].
#    -a, --aggregate     Aggregates npp/n20 ATMS granules into swaths prior
#                        to processing; you ought do this (to counter cross-
#                        granule effects) unless files are already aggregated.
#                        [default: --agg; --noagg is the contrary option].
#
# END-HELP (N.B. This tag is used in printing help, do not touch!)
#
# The following option is inactive; %options{processors} defaults to 1 and is
# never interrogated. Leave here in case I ever add multicore functionality.
#
#    -p PROCESSORS, --processors=PROCESSORS
#                        Number of cpus to use for MIRS processing.
#
# NOTES-TO-SELF:
# -------------
#   $!=N to change exit code on die() to N; default is 255
#
#   --clean does not run MIRS step_clean, but achieves some of the same
#   objectives by removing the temporary directory tree. If the user
#   selects --noclean the temporary directory is NOT removed and all MIRS
#   product files are available in the temporary directory structure.
#   This option might be useful for debugging and is necessary if the
#   user wants access to product files generated prior to the output
#   netCDF files that are copied up to WORK_DIR.  It becomes fully the
#   user's responsibility dispense with any unwanted directories/files.
#
#
# This is a valid way to run a single S-NPP granule ...
#
#   ./run_mirs.pl -s npp \
#     -i "/dataDir/{GATMO,TATMS,SATMS}_npp_d20131120_t0656260_*_cspp_dev.h5"
#
# The quotes are to prevent OS expanding wildcards/tokens on the commandline.
#
# For Metop-A files with NOAA filenaming convention, something like this ...
#
#   ./run_mirs.pl -x -s metopA \
#     -i "/dataDir/NSS.{AMA,MHS}X.M2.D13329.S0346.E0529.B3684546.SV"
#
# ... will run a single overpass or orbit file.
#
# But it is a whole lot simpler to put your data into a directory and to
# process it all with something like (a DB format NOAA-18 example now) ...
#
#   ./run_mirs.pl -s n18 -i /dataDir
#


# Pragmata
# ````````
use 5.006;
use strict;
use warnings;


# Load Perl modules - standard modules only
# `````````````````````````````````````````
use POSIX;
use File::Spec;
use File::Copy;
use Data::Dumper;
use Getopt::Long;
use File::Basename;
use Text::ParseWords;
use Encode qw(decode);
use Cwd qw(abs_path getcwd);
use Time::gmtime qw(:FIELDS);
use List::Util qw(sum min max);
use File::Path qw(mkpath rmtree);
use Time::Local qw(timegm_nocheck);
use File::Temp qw(tempfile tempdir);
use Scalar::Util qw(looks_like_number);
use I18N::Langinfo qw(langinfo CODESET);


# Global variables are all UPPERCASE throughout
# `````````````````````````````````````````````
our $AGG_WINDOW_MINUTES = 30; # 30-minute aggregation window for npp/n20
our $SLEEP_GFS_MINUTES = 0;   # 0 means disable wait for preferred GFS data
our $MAIN = basename($0);
defined $ENV{CSPP_MIRS_HOME} or die "$MAIN\:: CSPP_MIRS_HOME not set.\nAbort";
defined $ENV{CSPP_MIRS_REVNAME} or die "$MAIN\:: CSPP_MIRS_REVNAME not set.\nAbort";
our $HOME = File::Spec -> catfile($ENV{CSPP_MIRS_HOME}, $ENV{CSPP_MIRS_REVNAME});
(-d $HOME) or die "$MAIN\:: CSPP_MIRS subdirectory $HOME not found.\nAbort";
our $LATEST = 'working'; # root of temporary directory tree for dynamic data 


# $H5DUMP points to supplied static-compiled h5dump (version 1.8.11)
# ``````````````````````````````````````````````````````````````````
our $H5DUMP = File::Spec -> catfile($ENV{'CSPP_MIRS_HOME'}, 'bin', 'h5dump');
(-e $H5DUMP) or die "$MAIN\:: h5dump $H5DUMP not found.\nAbort";
(-x $H5DUMP) or die "$MAIN\:: h5dump $H5DUMP not executable.\nAbort";


# $H5REPACK points to supplied static-compiled h5repack
# `````````````````````````````````````````````````````
# But you can point to your own h5repack if you wish.
our $H5REPACK = File::Spec -> catfile($ENV{'CSPP_MIRS_HOME'}, 'bin', 'h5repack');
(-e $H5REPACK) or die "$MAIN\:: h5repack $H5REPACK not found.\nAbort";
(-x $H5REPACK) or die "$MAIN\:: h5repack $H5REPACK not executable.\nAbort";


# $H5NAGG points to supplied static-compiled nagg
# ```````````````````````````````````````````````
# But you can point to your own nagg if you wish. Ours is now 1.6.2-
our $H5NAGG = File::Spec -> catfile($ENV{'CSPP_MIRS_HOME'}, 'bin', 'nagg');
(-e $H5NAGG) or die "$MAIN\:: nagg $H5NAGG not found.\nAbort";
(-x $H5NAGG) or die "$MAIN\:: nagg $H5NAGG not executable.\nAbort";


# $WGRIB2 points to supplied static-compiled wgrib2
# `````````````````````````````````````````````````
# But you can point to your own wgrib2 if you wish
our $WGRIB2 = File::Spec -> catfile($ENV{'CSPP_MIRS_HOME'}, 'bin', 'wgrib2');
(-e $WGRIB2) or die "$MAIN\:: wgrib2 $WGRIB2 not found.\nAbort";
(-x $WGRIB2) or die "$MAIN\:: wgrib2 $WGRIB2 not executable.\nAbort";


# The MIRS system uses just one file to specify a file group
# ``````````````````````````````````````````````````````````
# Don't mess with these unless you are confident of your actions.
# There is no ambiguity between DB and DAAC for S-NPP filenames,
# so they are hard-coded (later) in the script.
our ($AMSUA_DB, $MHS_DB) = qw(amsua mhs);
our ($AMSUA_NOAA, $MHS_NOAA) = qw(AMAX MHSX);


# Get the command line options, display help or revision if asked
# ```````````````````````````````````````````````````````````````
my $revision = "CSPP_MIRS v2.1 $ENV{CSPP_MIRS_REVDATE}";
my $pwd = getcwd();
my %options = get_commandline($revision, $pwd);


# At present we are saying that lftp is a system requirement
# ``````````````````````````````````````````````````````````
our $LFTP = 'lftp';
($options{quarantine}) or do {
  ($options{SFR}) and do {
    system("$LFTP --version 1>/dev/null 2>/dev/null") == 0
      or die "$MAIN\:: Required utility lftp not found\nAbort"; }; };


# These are used for fileglobbing when INPUTFILES is a directory name
# ```````````````````````````````````````````````````````````````````
# Change to suit your site; for more control edit function validate_options(),
# or provide "<dirname>/<my_filemask>" on the command line as INPUTFILES; the
# "quotes" are important to prevent the OS expanding wildcard tokens.
my $group = "{$AMSUA_DB,$MHS_DB}";
my %filemask = (npp => '{GATMO,SATMS,TATMS}_npp*.h5',
                n18 => $group.'l1b_noaa18*.l1b',
                n19 => $group.'l1b_noaa19*.l1b',
                n20 => '{GATMO,SATMS,TATMS}_j01*.h5',
                metopA => $group.'l1b_M02*.l1b',
                metopB => $group.'l1b_M01*.l1b');
($options{xfiles}) and do {
  $group = "{$AMSUA_NOAA,$MHS_NOAA}";
  %filemask = (npp => '{GATMO,SATMS,TATMS}_npp*.h5',
               n18 => '*.'.$group.'.NN*',
               n19 => '*.'.$group.'.NP*',
               n20 => '{GATMO,SATMS,TATMS}_j01*.h5',
               metopA => '*.'.$group.'.M2*',
               metopB => '*.'.$group.'.M1*'); };
($options{verbose} == 3) and pretty_dump(%filemask);


# Define dynamic data remote & local locations and file naming conventions
# ````````````````````````````````````````````````````````````````````````
# The root of the local directory for dynamic ancillary data is either defined
# on the command line or is assigned a default value. I am appending the agreed
# subdirectory naming convention to this root and copying it across to a hash
# that contains much of what I need to know to fetch and store ancillary data.
# See https://jpssdb.ssec.wisc.edu/cspp_v_2_0/ancillary/ for directory structure.
# In this application the subdirectory structure '%Y_%m_%d_%j' is effectively
# hard-coded to match ancillary server, it is computed in make_GFS2BIN_commands().
my $local_repo = File::Spec -> catfile($options{dynanc}, '%Y_%m_%d_%j');
my $grib2_template = 'gfs.t%Hz.pgrb2.0p50.f0%f.%Y%m%d';
my %dynanc_loc = (GFSPROTO => 'https://',
                  GFSSITE => 'jpssdb.ssec.wisc.edu',
                  GFSREMOTE => 'cspp_v_2_0/ancillary/%Y_%m_%d_%j',
                  GFSTEMPLATE => $grib2_template,
                  GFSLOCAL => $local_repo);

# Parse and validate command line %options & make temporary working directory
# ```````````````````````````````````````````````````````````````````````````
%options = validate_options($filemask{$options{sat}}, %options);
my $tmpdir = tempdir('TMP_XXXXXX', DIR => $options{work}, CLEANUP => $options{clean});
chdir($tmpdir) or
  die "$MAIN\:: Cannot chdir to temporary working directory $tmpdir\nAbort";
($options{verbose} >= 2) and
  print "chdir to working directory $tmpdir\n";


# Make PCF and SCS scripts for this runtime environment
# `````````````````````````````````````````````````````
my $pcf = make_PCF($tmpdir, %options);
my $scs = make_SCS($tmpdir, $pcf, %options);
($options{verbose} >= 2) and do {
  print "PCF file is:  $pcf\n";
  print "SCS file is:  $scs\n"; };


# IF SnowFall Rate (-f, --SFR) determine which GFS to BIN converter to use
# ````````````````````````````````````````````````````````````````````````
our $GFSDATA = File::Spec -> catfile('input', 'GFS');
our $GFS2BIN = make_GFS2BIN_script(%options);
($options{SFR}) and do {
  mkpath($GFSDATA) or die "$MAIN\:: GFS directory $GFSDATA cannot be made.\nAbort";
  ($options{verbose} >= 2) and do {
    print "GFS2BIN  is:  $GFS2BIN\n";
    print "GFSDATA  is:  $GFSDATA\n"; }; };


# Preprocess and stage data for MIRS system compatibility
# ``````````````````````````````````````````````````````
%options = cspp_preprocess(%options);
($options{SFR}) or do {
  ($options{verbose} == 3) and pretty_dump(%options); };


# Determine dates/times needed for GFS to support SnowFall Rate & fetch
# `````````````````````````````````````````````````````````````````````
# I can set up the GFS files and links to them in the temporary working
# directory now, but I must wait for each overpass directory to be spawned
# in exec_SCS() before converting to binary the specific files I need
# because the naming convention of the .bin files is not sufficiently
# unique - files created for different overpasses can clobber each other.
($options{SFR}) and do {
  # After this, for each @{$options{process_files}} there will be a matching
  # command in @{$options{gfs2bin}} for converting GFS files.
  %options = make_GFS2BIN_commands(%options);
  ($options{verbose} == 3) and pretty_dump(%options); 
  # Collapse down to unique $YYYY$MM$DD$HH to avoid duplicate requests
  my @tmp = (); 
  for (my $i=0; $i<=$#{$options{process_files}}; $i++) {
    my @parts = split(/\s+/, $options{gfs2bin}[$i]);
    my $key = join('', $parts[2], $parts[1]);
    push (@tmp, $key);
  }
  # Assemble a unique list of YYYYMMDDHHs before fetching GFS
  my @YYYYMMDDHH = do { my %seen; grep { !$seen{$_}++ } @tmp };
  # Fetch GFS and wrangle into place for all dates to be processed
  populate_dynanc_gfs(\%dynanc_loc, \%options, @YYYYMMDDHH);
};


# If SFR and amsua_mhs instrument, I need a workaround for sfr_amsua
# ``````````````````````````````````````````````````````````````````
# File "../data/StaticData/SFR/limit_prod_amsua_mhs.dat" is baked into
# the sfr_amsua executable.  My workaround is to make a link to the mirs
# /data/ directory as $tmpdir/data/.  I will need to cd into a level down
# (say working/) before I run sfr_amsua. I modify cspp_<sat>_scs.bash
# and script_functions.bash (see lines 3456/7 of script_functions.bash)
# to make this work.
($options{SFR}) and do {
  (grep {$options{sat} eq $_} qw/n18 n19 metopA metopB/) and do {
    my $target = File::Spec -> catfile($HOME, 'data');
    symlink($target, 'data') or warn "$MAIN\:: WARNING: Cannot link to $target";
    my $found_start = ' sfr ..satId....sensor1....fmsdrDir';
    my $line1 = '  sfr ${satId} ${sensor1} ${fmsdrDir#*/} ${extResol} $(basename ${fmsdr4SfrList}) ${sfrDir#*/} ${sfrParam}\\';
    my $line2 = '      ${sfrMask} ${sfrLatBox} ${sfrCoef} ${sfrLut} ../${logFile} ${sfrSrc} ${makeOrNot} ${binPath}\\';
    my $line3 = '      ${processMode} ${orbitInfo} ${retrOnOrbitOrSubOrbit} ${fmsdrChoppDir#*/}\\';
    my $line4 = '      ${nOrbits2Process} $(basename ${sfrControlFile}) ${nwpSfrGridPath#*/} $(basename ${gridNwpSfrList})\\';
    my $line5 = '      ${nProfs2Retr} ${sfr2depSrc} $(basename ${sfr2depControlFile}) $(basename ${sfrList}) ${depDir#*/} $(basename ${depList})';
    # Hoover up the complete $scs
    open (my $fh, "<$scs"); chomp (my @scs = <$fh>); close ($fh);
    my @script = ();
    my $skipLines = 0;
    foreach (@scs) {
      (/$found_start/) and do {
        # found the marker text, replace lines 1-5 above
        push (@script, '  cd working', $line1, $line2, $line3, $line4, $line5, '  cd ..');
        $skipLines = 5;
      };
      ($skipLines) or push (@script, $_);
      ($skipLines) and $skipLines--;
    }
    # Write the modified $scs
    open (my $FH, ">$scs");
    foreach (@script) { print $FH "$_\n"; }
    close ($FH);
  };
};


# Run the job (worry later about specifying nCPUs)
# ````````````````````````````````````````````````
exec_SCS($scs, %options); # internal abort on error, $tmpdir sub-tree also made here


# Leave a spare directory tree - useful for debugging
# ```````````````````````````````````````````````````
($options{clean}) or make_tmpdir_tree($options{sat});


# All done - chdir so that temporary working directory can be disappeared
# ```````````````````````````````````````````````````````````````````````
chdir($pwd) or die "$MAIN\:: Cannot chdir back to $pwd\nAbort";
exit(0);



###############################################################################
#                                                                             #
#         User-defined functions - they are listed alphabetically             #
#                                                                             #
#   NOTE: My editor can collapse code between matching {} braces, this        #
#   explains the strange formatting - I can collapse each function to see     #
#   just the function name and an example of how to call it; and expand       #
#   to see the more detailed documentation and to edit.      JED Nov 2013     #
#                                                                             #
###############################################################################

# -------------------------------
  sub convert_YYYYJJJ_to_YYYYMMDD
# ===============================
# $YYYYMMDD = convert_YYYYJJJ_to_YYYYMMDD($YYYYJJJ)
  {
    # No overloading, input is scalar string, expects to use
    # Time::gmtime qw(:FIELDS) & Time::Local qw(timegm_nocheck)
    my $YYYYJJJ = shift;
    # I don't need this complexity, but I want to keep the concept in vogue,
    # and support sub token_value() in use, because it is needed elsewhere.
    my $mask = '%Y%j';
    my $YYYY = token_value('%Y', $YYYYJJJ, $mask);
    my $JJJ  = token_value('%j', $YYYYJJJ, $mask);
    gmtime(timegm_nocheck(0, 0, 0, $JJJ, 0, $YYYY));
    my $MM = sprintf("%02d", $tm_mon + 1);
    my $DD = sprintf("%02d", $tm_mday);
    return "$YYYY$MM$DD";
  }

# -------------------------------
  sub convert_YYYYMMDD_to_YYYYJJJ
# ===============================
# $YYYYJJJ = convert_YYYYMMDD_to_YYYYJJJ($YYYYMMDD)
  {
    # No overloading, input is scalar string, expects to use
    # Time::gmtime qw(:FIELDS) & Time::Local qw(timegm_nocheck)
    my $YYYYMMDD = shift;
    # I don't need this complexity, but I want to keep the concept in vogue,
    # and support sub token_value() in use, because it is needed elsewhere.
    my $mask = '%Y%m%d';
    my $YYYY = token_value('%Y', $YYYYMMDD, $mask);
    my $MM   = token_value('%m', $YYYYMMDD, $mask);
    my $DD   = token_value('%d', $YYYYMMDD, $mask);
    gmtime(timegm_nocheck(0, 0, 0, $DD, $MM-1, $YYYY));
    my $JJJ = sprintf("%03d", $tm_yday + 1);
    return "$YYYY$JJJ";
  }

# -------------------
  sub cspp_preprocess
# ===================
# %options = cspp_preprocess(%options);
  {
    # On input, %options comprises some command line parameters plus other
    # processing options, including defaults for unspecified command line values
    # and extra stuff defined in main to guide processing. On output, %options
    # needs to include a field $options{process_files} that points to files
    # that satisfy MIRS system requirements in terms of both filename and
    # internal format.  NOTE:  I can find no evidence that process_files is
    # used, but it is useful as list of one file per file group (i.e. for
    # npp/n20 it is just TATMS out of GATMO,SATMS,TATMS and for n18/19 and
    # metopA/B just AMSUA out of AMSUA,MHS). So I amgoign to heep it and use
    # for guiding the GFS data needed for SnowFall Rate.
    # My old nagg command, for joining npp granules, was :
    # my $cmd = "$H5NAGG -S -t TATMS,SATMS -n 150 -d . $granules/*_npp_*.h5";
    # This sometimes caused pre-processsing aggregation into two files per
    # swath instead of one because of an enforced break at some time or partial
    # orbit boundary.  The newer versions of nagg have a --onefile option to
    # deal with this occurence. NOTE: from CSPP_MIRS_2_0 I will need to update
    # nagg to one that supports j01 - DONE - but beware that this is presently
    # broken for some file types, just not ones used in this application.
    # UPDATE:  I am using a dynamic-linked but older version of nagg that has
    # fewer lib dependencies because I am concerned about lib requirements
    # clashing between the j01 supported nagg and the MIRS dynamic executables.
    # This will mean that preaggregated files will have _npp_ in their names
    # but they will exist only in the temporary working directory and it seems
    # that there are no deleterious side effects.             JED  03-Aug-2018
    my (%options) = @_;
    my $name = (caller(0))[3];
    $name =~ s/main/$MAIN/g;
    my @process_files = ();
    (grep {$options{sat} eq $_} qw/npp n20/) and do {
      if ($options{aggregate}) {
        my %swaths = npp_granules_to_swaths(@{$options{input_files}});
        my $iswath = 0;
        foreach my $key (sort keys %swaths) {
          my $granules = sprintf("granules_%03d", ++$iswath);
          mkpath($granules) or die "$name\: WORK_DIR $granules cannot be made.\nAbort";
          foreach (@{$swaths{$key}}) {
            my $source = File::Spec -> catfile($options{input_dir}, $_);
            my $dest = File::Spec -> catfile($granules, $_);
            my $cmd = "$H5REPACK -n -f NONE $source $dest";
            system($cmd) == 0 or die "$name\: Could not run command $cmd\nAbort";
          }
          my $cmd = "$H5NAGG -S -t TATMS,SATMS --onefile -d . $granules/*_???_*.h5";
          ($options{verbose} < 2) and $cmd = "$cmd 1>/dev/null 2>/dev/null";
          system($cmd) == 0 or die "$name\: Could not run command $cmd\nAbort";
          #
          # The version of nagg I am using is static (good) but creates files with
          # _npp_ in filename even for _j01_ input files. So I am going to fix that
          # here and then check if there are any unwanted downstream effects.  The
          # alternative is to use the dynamic-linked version of nagg that I have,
          # but I have a concern that I end up with a conflicting set of dynamic
          # library requirements to support both nagg and the MIRS executables,
          # the latter also created as dynamic-linked.             JED  26-Jul-2018
          #
          ($options{sat} eq 'n20') and do {
            foreach (glob('*_npp_*')) {
              my $target = $_;
              $target =~ s/_npp_/_j01_/g;
              move($_, $target) or die "$name\: Cannot rename $_\nAbort";
            }
          };
        }
      } else {
        foreach (@{$options{input_files}}) {
          my $source = File::Spec -> catfile($options{input_dir}, $_);
          my $cmd = "$H5REPACK -n -f NONE $source $_";
          system($cmd) == 0 or die "$name\: Could not run command $cmd\nAbort";
        }
      }
      my @gatmo = sort(glob('GATMO*'));
      my @satms = sort(glob('SATMS*'));
      my @tatms = sort(glob('TATMS*'));
      ($#tatms == $#gatmo) or die "$name\: Number of TATMS != GATMO\nAbort";
      ($#satms == $#tatms) or die "$name\: Number of SATMS != TATMS\nAbort";
      @process_files = @tatms; };
    (grep {$options{sat} eq $_} qw/n18 n19 metopA metopB/) and do {
      foreach (@{$options{input_files}}) {
        my $pf = undef;
        my $df = File::Spec -> catfile($options{input_dir}, $_);
        (index($_, $AMSUA_DB) != -1 or index($_, $AMSUA_NOAA) != -1) and do {
          $pf = make_L1B_bigEndian('amsua', $df); };
        (index($_, $MHS_DB) != -1 or index($_, $MHS_NOAA) != -1) and do {
          $pf = make_L1B_bigEndian('mhs', $df); };
        defined($pf) or die "$name\: Could not endian convert $_\nAbort";
        $options{site_id} = substr($pf, 0, 3);
        my $mf = 'NSS'.substr($pf, 3);
        $mf =~ s/\.MB\./\.M1\./; $mf =~ s/\.MA\./\.M2\./;
         ($mf eq $pf) or do { # avoid link-to-self (which will fail!)
          symlink($pf, $mf) or die "$name\: Cannot link $pf\nAbort"; };
        (index($pf, $AMSUA_NOAA) != -1) and push(@process_files, $mf);
      } 
      my @amsua = sort(glob("NSS.$AMSUA_NOAA.*"));
      my @mhs = sort(glob("NSS.$MHS_NOAA.*"));
      my $n = $#amsua;
      ($n == $#mhs) or die "$name\: Number of $AMSUA_NOAA != $MHS_NOAA\nAbort";
      for (my $i=0; $i<=$n; $i++) {
        my ($a1, $Sa, $a2, $Ea, $a3) = unpack('a20a4a2a4', $amsua[$i]);
        (looks_like_number($Sa) and looks_like_number($Ea)) or
          die  "$name\: $options{sat} $amsua[$i] unrecognized\nAbort";
        my ($m1, $Sm, $m2, $Em, $m3) = unpack('a20a4a2a4a12', $mhs[$i]);
        (looks_like_number($Sm) and looks_like_number($Em)) or
          die  "$name\: $options{sat} $mhs[$i] unrecognized\nAbort";
        ((abs($Sa-$Sm) + abs($Ea-$Em)) > 0) and do {
          ($options{verbose} > 1) and 
            warn "$name\: WARNING! $AMSUA_NOAA:$MHS_NOAA mismatch. Forcing\n";
          my $forced = pack('a20a4a2a4a12', $m1, $Sa, $m2, $Ea, $m3);
          move("$mhs[$i]", "$forced"); };
      } };
    @{$options{process_files}} = @process_files;
    return %options;
  }

# ---------------------------
  sub datestring_for_filename
# ===========================
# my $datestring = datestring_for_filename($year, $month, $day, $UTC);
  {
    my ($year, $month, $day, $UTC) = (shift, shift, shift, shift);
    my $hour = int($UTC/3600);
    my $minute = int($UTC/60 - $hour*60);
    my $tsec = int(($UTC - $hour*3600 - $minute*60) * 10);
    my $fmt = "%04d%02d%02d%02d%02d%03d";
    my $datestring = sprintf($fmt, $year, $month, $day, $hour, $minute, $tsec);
    return $datestring;
  }

# ------------
  sub exec_SCS
# ============
# exec_SCS($SCS, %options);
  {
    # $SCS is the runtime SCS file created by make_SCS()
    # %options contains things I need like the SDR/L1B files to
    # process plus verbosity and do-nothing choices. I do not need
    # the name of the PCF because it is already named in the SCS.
    my $SCS = shift;
    my (%options) = @_;
    my $name = (caller(0))[3];
    $name =~ s/main/$MAIN/g;
    ($options{nothing} and ($options{verbose} > 0)) and
      print "$name\: Do not execute mode --nothing\n";
    for (my $i=0; $i<=$#{$options{process_files}}; $i++) {
      my $parked = sprintf("swath_%03d", ($i+1)); # make it 1-based
      my $outdir = make_tmpdir_tree($options{sat});
      my $cmd = $options{gfs2bin}[$i];
      ($options{verbose} < 3) and $cmd = "$cmd 1>/dev/null 2>/dev/null";
      ($options{verbose} > 1) and print "$cmd\n";
      system($cmd)==0 or die "$name\: Cannot complete $cmd\nAbort";
      $cmd = sprintf("%s %s", $SCS, $options{process_files}[$i]);
      ($options{verbose} < 1) and $cmd = "$cmd 1>/dev/null 2>/dev/null";
      ($options{nothing}) or do {
        system($cmd)==0 or die "$name\: Cannot complete $cmd\nAbort"; };
      my $loglobber = '*logFile*.dat';
      foreach (glob($loglobber)) {
        copy($_, $options{work}) or die "$name\: Copy failed: $!\nAbort";
      }
      my $ncglobber = File::Spec -> catfile($outdir, 'NPR-MIRS*.nc');
      foreach (glob($ncglobber)) {
        copy($_, $options{work}) or die "$name\: Copy failed: $!\nAbort";
        (grep {$_ eq $options{sat}} qw/n18 n19 metopA metopB/) and do {
          my $oldname = File::Spec -> catfile($options{work}, basename($_));
          my $newname = File::Spec -> catfile($options{work},
            basename(verified_MIRS_filename($_)));
          rename ($oldname, $newname) or die "$name\: Rename failed: $!\nAbort";
          ($options{verbose} >= 1) and do {
            (filenames_differ_significantly($oldname, $newname)) and do {
              my $old = basename($oldname);
              my $new = basename($newname);
              print "  Changed: $old\n";
              print "       to: $new\n";
            };
          };
        };
      }
      move("$LATEST", "$parked");
    }
    return;
  }

# ----------------------------------
  sub filenames_differ_significantly
# ==================================
# my $status = filenames_differ_significantly($oldname, $newname);
  {
    # The assumption is that basename($newname) is good.
    # It should be like ...
    # NPR-MIRS-IMG_v11r3_n19_s201901072259000_e201901072310000_c201901141955130.nc
    # ... or ...
    # NPR-MIRS-SND_v11r3_n19_s201901072259000_e201901072310000_c201901141955130.nc
    # ... both of which are 76 characters long.
    # I will abort if the $new filename is not this length, and return 1 if 
    # $old file is not this length.
    # Otherwise I return 0 if year, month, dom, hour, minute, sec are the same for
    # start and end date/time (not creation), but will return 1 if not. So I am
    # not requiring seconds (actually tenth of seconds) to be the same, MIRS zero
    # fills the filename here for some missions.
    my ($oldname, $newname) = (shift, shift);
    my $name = (caller(0))[3];
    $name =~ s/main/$MAIN/g;
    my $old = basename($oldname);
    my $new = basename($newname);
    (length($new) == 76) or die "$name\: Bad MIRS output filename $new\nAbort";
    (length($old) == 76) or return 1;
    my $newstring = substr($new, 23, 16).substr($new, 40, 16);
    my $oldstring = substr($old, 23, 16).substr($old, 40, 16);
    ($oldstring eq $newstring) or return 1;
    return 0;
  }

# ------------------
  sub first_and_last
# ==================
# my ($first, $last) = first_and_last($filename, $field);
  {
    my ($file, $var) = (shift, shift);
    my $name = (caller(0))[3];
    $name =~ s/main/$MAIN/g;
    my $tmp = '';
    my $cmd = "$H5DUMP -d $var $file |";
    open(my $fh, $cmd) or die "$name\: Could not run command $cmd\nAbort";
    while (<$fh>) { chomp;
      /   DATASPACE  SIMPLE \{/ and do {
        while (<$fh>) { chomp;
          /   DATA \{/ and do {
            while (<$fh>) { chomp;
              /   \}/ and last;
              my @parts = split(/\:/, $_);
              $tmp = $tmp.$parts[1];
            }
            goto DONE;
          };
        }
      };
    }
    DONE:
    close($fh);
    my @parts = split(/\,/, $tmp);
    my $first = trim_string($parts[0]);
    my $last = trim_string($parts[$#parts]);
    return ($first, $last);
  }

# -------------------
  sub get_commandline
# ===================
# %options = get_commandline($revision, $pwd);
  {
    # If things go bad in here, jump to usage() which terminates with exit(9).
    # Note that I had to hack in the Python-like -vv & -vvv support.
    # This routine assumes standard module Getopt::Long has been loaded.
    # abs_path() is dangerous for the purpose of setting the work (and,
    # equivalently, the output) directory - see this post:
    # http://bugs.debian.org/cgi-bin/bugreport.cgi?bug=257568
    my ($rev, $pwd) = (shift, shift);
    my $name = (caller(0))[3];
    $name =~ s/main/$MAIN/g;
    my $codeset = langinfo(CODESET);
    @ARGV = map{decode $codeset, $_} @ARGV;
    foreach (@ARGV) { s/\?/-\?/g; }
    my %options = ();
    GetOptions(\%options,
      'version|V',
      'help|?',
      'nothing|n',
      'input_files|i=s',
      'sat|s=s',
      'work|w=s',
      'dynanc|d=s',
      'resolution|r=s',
      'block|b=i',
      'processors|p=i',
      'verbose|v+', 'vv', 'vvv',
      'SFR|f!',
      'xfiles|x',
      'aggregate|a!',
      'quarantine|q!',
      'clean|c!') or usage();
    defined($options{version}) and die "$rev\n";
    defined($options{help}) and usage();
    $options{nothing} = defined($options{nothing}) ? $options{nothing} : 0;
    $options{xfiles} = defined($options{xfiles}) ? $options{xfiles} : 0;
    defined($options{input_files}) or usage();
    defined($options{sat}) or usage();
    (grep {$_ eq $options{sat}} qw/npp n18 n19 n20 metopA metopB/) or usage();
    (defined($options{work})) or $options{work} = $pwd;
    (defined($options{resolution})) or $options{resolution} = 'HI';
    (defined($options{block})) or $options{block} = 2;
    (defined($options{dynanc})) or 
      $options{dynanc} = File::Spec -> catfile($ENV{CSPP_MIRS_HOME}, 
        'data', 'dynamic_ancillary');
    foreach (qw/dynanc work/) {
      (substr($options{$_}, 0, 1) eq '/') or do {
        $options{$_} = File::Spec -> catfile($pwd, $options{$_}); };
    }
    (defined($options{processors})) or $options{processors} = 1;
    $options{verbose} = defined($options{verbose}) ? $options{verbose} : 0;
    $options{vv} = defined($options{vv}) ? $options{vv} : 0;
    $options{vvv} = defined($options{vvv}) ? $options{vvv} : 0;
    $options{verbose} = $options{verbose} + 2 * $options{vv};
    $options{verbose} = $options{verbose} + 3 * $options{vvv};
    $options{verbose} = min($options{verbose}, 3);
    delete @options{'vv', 'vvv'};
    $options{SFR} = defined($options{SFR}) ? $options{SFR} : 1;
    $options{clean} = defined($options{clean}) ? $options{clean} : 1;
    $options{aggregate} = defined($options{aggregate}) ? $options{aggregate} : 1;
    $options{quarantine} = defined($options{quarantine}) ? $options{quarantine} : 0;
    return %options;
  }

# -----------------------------------------------------------
  sub is_integer { defined $_[0] and $_[0] =~ /^[+-]?\d+$/; }
# ===========================================================

# -------------------------
  sub is_plausible_datetime
# =========================
# $status = is_plausible_date($YYYY, $MM, $DD, $hh, $mm);
  {
    # NOTE: Year < 2000 is considered implausible in this function.
    my ($YYYY, $MM, $DD, $hh, $mm) = (@_);
    my $implausible_year = 2000;
    is_integer($YYYY) or return 0;
    $YYYY < $implausible_year and return 0;
    is_integer($MM) or return 0;
    $MM >= 1 and $MM <= 12 or return 0;
    is_integer($DD) or return 0;
    $DD >= 1 and $DD <= 31 or return 0;
    is_integer($hh) or return 0;
    $hh >= 0 and $hh <= 23 or return 0;
    is_integer($mm) or return 0;
    $mm >= 0 and $mm <= 59 or return 0;
    return 1;
  }

# ---------------
  sub lftp_mirror
# ===============
# $status = lftp_mirror($site, $file, $local, %options);
  {
    # executes lftp in mirror mode, but if NOT $options{quarantine}
    # added flock to download directory on 08-Jan-2019 to prevent collisions
    my ($site, $file, $local) = (shift, shift, shift);  
    my (%options) = @_;
    my $name = (caller(0))[3];
    $name =~ s/main/$MAIN/g;
    my $v = '';
    ($options{verbose} == 3) and $v = '-v';
    -e "$local/$file" and do {
      ($options{verbose} == 3) and print "$name\: $file FOUND in local repo\n";
      return 1;
    };
    ($options{quarantine}) and do {
      ($options{verbose} >= 1) and 
        warn "$name\: WARNING: $file NOT FOUND in local repo AND --quarantine";
      return 0;
    };
    (-d $local) or do {
      ($options{verbose} >=2) and print "Making $local\n";
      mkpath($local) or
        die "$name\: $local cannot be made.\nAbort"; };
    open(my $fh, ">script_file.lftp") or die "$name\: Cannot make script_file.lftp $!";
    print $fh "set dns:order inet inet6\n";            # default is 'inet6 inet'
    print $fh "set ssl:verify-certificate no\n"; 
    #print $fh "set net:timeout 1\n";                   # default is '5'
    #print $fh "set net:max-retries 5\n";               # default is '10'
    #print $fh "set net:persist-retries 1\n";           # default is '5'
    #print $fh "set net:reconnect-interval-base 6\n";   # default is '60'
    print $fh "mirror $v -L -n --no-empty-dirs $site -I $file $local\n";
    close($fh);
    my $cmd = "flock $local $LFTP -f script_file.lftp";
    system($cmd) == 0 or do {
      ($options{verbose} >= 1) and warn "$name\: WARNING: $cmd failed";
      return 0; };
    return 1;
  }

# -------------------------
  sub make_GFS2BIN_commands
# =========================
#  %options = make_GFS2BIN_commands(%options);
  {
    # After this, for each @{$options{process_files}} there will be a 
    # matching command in @{$options{gfs2bin}} for converting GFS files.
    my (%options) = @_;
    my $name = (caller(0))[3];
    $name =~ s/main/$MAIN/g;
    @{$options{gfs2bin}} =();
    my ($YYYYMMDD, $HH) = undef;
    foreach (@{$options{process_files}}) {
      (grep {$options{sat} eq $_} qw/npp n20/) and do {
        my $mask = 'TATMS_sat_d%Y%m%d_t%H%M%SX_e*';
        my $YYYY = token_value('%Y', $_, $mask);
        my $MM = token_value('%m', $_, $mask);
        my $DD = token_value('%d', $_, $mask);
        $HH = token_value('%H', $_, $mask);
        $YYYYMMDD = "$YYYY$MM$DD";
      };
      (grep {$options{sat} eq $_} qw/n18 n19 metopA metopB/) and do {
        my $mask = 'NSS.AMAX.NN.D%y%j.S%H%M.E*';
        # n18 started in 2005, so the following is safe
        my $YYYY = '20'.token_value('%y', $_, $mask);
        my $JJJ = token_value('%j', $_, $mask);
        $HH = token_value('%H', $_, $mask);
        $YYYYMMDD = convert_YYYYJJJ_to_YYYYMMDD("$YYYY$JJJ")
      };
      $HH = sprintf("%02d", $HH - ($HH % 6));
      push(@{$options{gfs2bin}}, "./$GFS2BIN $HH $YYYYMMDD");
    }
    return %options;
  }

# -----------------------
  sub make_GFS2BIN_script
# =======================
# our $GFS2BIN = make_GFS2BIN_script(%options);
  {
    # There will be an associated PCF in setup that will need to be edited and
    # sourced.  But it is also true that the GFS filemask used in $GFS2BIN may
    # not match the filemask used on the ancilary data server so that I will need
    # to either make bunch of links or make some on-the-fly changes to a copy of
    # $GFS2BIN put into the temporary working directory.
    # BREAKING NEWS: Kathy will stage as gfs.t06z.pgrb2.0p50.f012.20180415
    # so I do not need to make any on-the-fly name changes.  JED 20-Jul-2018
    my (%options) = @_;
    my $name = (caller(0))[3];
    $name =~ s/main/$MAIN/g;
    my $gfs2bin = undef;
    my $OutputBinDataDir = File::Spec -> catfile($LATEST, 'ExternalData', 'gridNWP_sfr');
    (grep {$options{sat} eq $_} qw/npp n20/) and do {
      $gfs2bin = File::Spec -> catfile($HOME, 'scripts', 'gfs_atms_sfr.bash');
      $OutputBinDataDir = File::Spec -> catfile($OutputBinDataDir, 'atms'); };
    (grep {$options{sat} eq $_} qw/n18 n19 metopA metopB/) and do {
      $gfs2bin = File::Spec -> catfile($HOME, 'scripts', 'gfs_amsua_sfr.bash');
      $OutputBinDataDir = File::Spec -> catfile($OutputBinDataDir, 'amsua'); };
    (-e $gfs2bin) or die "$MAIN\:: SFR needs GFS converter $gfs2bin not found.\nAbort";
    (-x $gfs2bin) or die "$MAIN\:: SFR needs GFS converter $gfs2bin not executable.\nAbort";
    # Hoover up the complete $gfsbin
    open (my $fh, "<$gfs2bin"); chomp (my @gfs2bin = <$fh>); close ($fh);
    $gfs2bin = basename($gfs2bin);
    my $gfspcf = $gfs2bin; $gfspcf =~ s/sfr\.bash/sfr_pcf\.bash/g;
    $gfspcf = File::Spec -> catfile($HOME, 'setup', $gfspcf);
    (-e $gfspcf) or die "$MAIN\:: GFS converter $gfs2bin needs $gfspcf not found.\nAbort";
    # Hoover up the complete $gfspcf
    open ($fh, "<$gfspcf"); chomp (my @gfspcf = <$fh>); close ($fh);
    # Combine @gfs2bin & @gfspcf into a single script
    my @script = ();
    foreach (@gfs2bin) {
      (/\. \.\.\/setup\/gfs/) and do { 
        foreach (@gfspcf) {
          next if /^\s*(#.*)?$/;
          /WGRIB=/ and $_ = "WGRIB=$WGRIB2";
          /InputGribDataDir=/ and $_ = "InputGribDataDir=$GFSDATA";
          /OutputBinDataDir=/ and $_ = "OutputBinDataDir=$OutputBinDataDir";
          push (@script, $_);
        } next; };
      push (@script, $_);
    }
    # Write the localized GFS converter script
    open (my $FH, ">$gfs2bin");
    foreach (@script) { print $FH "$_\n"; }
    close ($FH);
    chmod (0755, $gfs2bin) or die "$name\: Cannot chmod $gfs2bin.\nAbort";
    return $gfs2bin;
  }

# ----------------------
  sub make_L1B_bigEndian
# ======================
# $outfile = make_L1B_bigEndian($instrument, $infile);
  {
    # $instrument is one of 'amsua' or 'mhs'
    # $infile is the name of the L1B file to test/convert endian
    # $outfile is the big endian output filename created
    #
    # $outfile is extracted from the header of $infile, bytes 23-64 inclusive
    # (42 bytes). The reason for creating $outfile in WORK_DIR/tempdir is so
    # that we don't clobber data files in WORK_DIR if the user happens to
    # specify WORK_DIR as the place where the original L1B files are stored.
    # For files named like "amsual1b_noaa18_20131014_2145_43295.l1b" no problem,
    # but for files named like "NSS.AMAX.NN.D06032.S0122.E0317.B0361819.WI",
    # problem since this is the same as the filename stored in its header!
    #
    # NOTE: We coalesce repeated character variables to strings in the $Lhdr and
    # $Lrec buffer specification (because we are not actually applying any logic
    # to characters), but only after 'a42a8' in $Lhdr, since the a42 field is to
    # be used to define $outfile.
    #
    my ($instrument, $infile) = (shift, shift);
    my $name = (caller(0))[3];
    $name =~ s/main/$MAIN/g;
    my ($Lhdr, $Bhdr, $Lrec, $Brec, $count, $buf) = undef;
    my ($reclen, $year, $iyear, $doy, $idoy, $iout) = undef;
    ($instrument eq 'amsua') and do {
      ($reclen, $iout, $iyear, $idoy) = (2560, 11, 22, 23);
      $Lhdr = 'a3av9a42a8v4Vv2V2v2Vv3v5Vv2V2v2Vv14Va16V4v106V115a8v8V363v6Vv2V46';
      $Lrec = 'v4Vv2V4v16V95v6V2v94V61v620Vv2V2a28Vv156Vv2V2a32v6V'; };
    ($instrument eq 'mhs') and do {
      ($reclen, $iout, $iyear, $idoy) = (3072, 11, 22, 23);
      $Lhdr = 'a3av9a42a8v4Vv2V2v2Vv8Vv2Vv15a16v51V53a8v8V92v436Vv2V6v642';
      $Lrec = 'v4Vv2VvaaV2v14V30v8V2v274V180v544V2v48V2a78v9V5a52v12Vv102'; };
    defined($reclen) or die "$name\: Instrument $instrument not supported\nAbort";
    $Bhdr = $Lhdr; $Bhdr =~ s/v/n/g; $Bhdr =~ s/V/N/g;
    $Brec = $Lrec; $Brec =~ s/v/n/g; $Brec =~ s/V/N/g;
    open my $fh, "<$infile"; binmode $fh;
    $count = read($fh, $buf, $reclen);
    ($count == $reclen) or die "$name\: Header count != reclen $infile\nAbort";
    my @data = unpack($Lhdr, $buf);
    ($year, $doy) = ($data[$iyear], $data[$idoy]);
    my $Lendian = ($year > 1960 and $year < 2099 and $doy > 0 and $doy < 367);
    @data = unpack($Bhdr, $buf);
    ($year, $doy) = ($data[$iyear], $data[$idoy]);
    my $Bendian = ($year > 1960 and $year < 2099 and $doy > 0 and $doy < 367);
    ($Lendian xor $Bendian) or die "$name\: $infile endian indeterminate\nAbort";
    my $outfile = $data[$iout];
    $outfile =~ s/^\s+//; $outfile =~ s/\s+$//; $outfile =~ s/ //g;
    (length($outfile) != 42) and die "$name\: L1B filename $outfile funky\nAbort";
    open my $FH, ">$outfile"; binmode $FH;
    ($Lendian) and $buf = pack($Bhdr, unpack($Lhdr, $buf));
    print $FH $buf;
    while (!eof($fh)) {
      $count = read($fh, $buf, $reclen);
      ($count == $reclen) or die "$name\: Record count != reclen $infile\nAbort";
      ($Lendian) and $buf = pack($Brec, unpack($Lrec, $buf));
      print $FH $buf;
    }
    close $fh; close $FH;
    return $outfile;
  }

# ------------
  sub make_PCF
# ============
# $pcf = make_PCF($tmpdir, %options);
  {
    # $tmpdir is fully-qualified pathname of temporary working directory.
    # %options are the command line options, parsed and validated.
    # $pcf is fully-qualified PCF filepath for this run.
    # $LATEST is name of the working directory within $tmpdir for the
    #   swath or granule currently being processed - really it is fixed
    #   in MAIN as 'working'.
    # $latest is the fully-qualified version of $LATEST
    #
    # The problem with $latest is that for some processes (at least for
    # RUNNING TDR->SDR(atms) STEP), the fully qualified filenames that are
    # prefixed by $latest get truncated to 200 characters in the MIRS system.
    # This can be a problem, not least because the filenames themselves can
    # be over 80 characters long. Since we have already chdir'd into $tmpdir,
    # we can use relative paths.
    #
    # Changes for CSPP_MIRS_2_0 arise from new functionality AND because I
    # have decided to NOT pre-edit any of the supplied PCF files in the
    # delivery.  The $CSPP_MIRS_REVNAME environment is now the dirname that
    # the DAP unpacks into (in CSPP_MIRS_1_0 I prefixed with cspp_) and I just
    # pick an existant PCF (i.e. as delivered) to use as a template.
    #
    my $tmpdir = shift;
    my (%options) = @_;
    my $name = (caller(0))[3];
    $name =~ s/main/$MAIN/g;
    my $tmp = File::Spec -> catfile($HOME, 'setup', $options{sat}.'_pcf_benchmark_hr.bash');
    -e $tmp or die "$MAIN\:: PCF template $tmp does not exist.\nAbort";
    my $pcf = File::Spec -> catfile($tmpdir, 'cspp_'.$options{sat}.'_pcf.bash');
    open(my $fhI, "<$tmp") or die "$name\: Cannot open $tmp for read.\nAbort";
    open(my $fhO, ">$pcf") or die "$name\: Cannot open $pcf for write.\nAbort";
    while (<$fhI>) { chomp;
      # text swaps from CSPP_MIRS_1_0
      (/date=/) and $_='date=dummy';
      (/rootPath=/) and $_='rootPath='.$HOME;
      (/rdrSensor1Path=/) and $_='rdrSensor1Path=./';
      (/rdrSensor2Path=/) and $_='rdrSensor2Path=./';
      (/rdrOrbitPath=/) and $_='rdrOrbitPath=./';
      (/controlDataPath=/) and $_='controlDataPath='.$LATEST.'/ControlData';
      (/inputDataPath=/) and $_='inputDataPath='.$LATEST.'/InputsData';
      (/externalDataPath=/) and $_='externalDataPath='.$LATEST.'/ExternalData';
      (/testbedDataPath=/) and $_='testbedDataPath='.$LATEST.'/TestbedData';
      (/step_clean=/) and $_='step_clean=0'; # MIRS cleaning not good for CSPP.
      (/useCPU=/) and $_='useCPU=0';
      (/step_choppRadFiles=/) and $_='step_choppRadFiles=0';
      (/nChoppedFilesPerOrbit=/) and $_='nChoppedFilesPerOrbit=1';
      # text swaps new for CSPP_MIRS_2_0
      (/logPath=/) and $_='logPath=./';
      (/step_mergeEdr=/) and $_='step_mergeEdr=0';
      (/retrOnOrbitOrSubOrbit=/) and $_='retrOnOrbitOrSubOrbit=0';
      (/outFMAccuracy=/) and $_='outFMAccuracy=0';
      (/fmType=/) and do { ($options{resolution} eq 'LO') and $_='fmType=0'; };
      # text swaps new for CSPP_MIRS_2_0 at patch level 2.0.2
      ($options{sat} eq 'npp') and do {
        ($options{block} == 1) and do {
          (/biasFileToUse=/) and $_='biasFileToUse=${biasPath}/biasCorrecBlock1_npp.dat';
        };
        # This was active for mirs_v11r1_r3636, but the bias files got cleaned up
        # in mirs_v11r3_r4112, so this is not necessary and will break system now.
        # This is change from CSPP_MIRS_2_0 at patch level 2.0.2 to CSPP_MIRS_2_1
        # I am goping to leave this here as a reminder.           JED  27-Jul-2018
        # ($options{block} == 2) and do {
        #   (/biasFileToUse=/) and do {
        #     $_='biasFileToUse=${biasPath}/biasCorrecBlock2_npp.dat';
        #     # in mirs_v11r3_r4112, biasCorrecBlock2_npp.dat is empty but
        #     # biasCorrec_npp.dat == biasCorrecBlock2_npp.dat of mirs_v11r1_r3636
        #     # so we make the substitution and this script will still work for
        #     # both mirs_v11r1_r3636 and mirs_v11r3_r4097.        JED 03-Jul-2018
        #     ((-s $_) < 75000) and $_='biasFileToUse=${biasPath}/biasCorrec_npp.dat';
        #   };
        # };
      };
      ($options{SFR}) or do { (/step_sfr=/) and $_='step_sfr=0'; };
      # text swaps new for CSPP_MIRS_2_2
      ($options{processors} > 1) and do {
        (/useCPU=/) and $_='useCPU=1';
        (/step_choppRadFiles=/) and $_='step_choppRadFiles=1';
        (/nChoppedFilesPerOrbit=/) and $_="nChoppedFilesPerOrbit=$options{processors}";
        (/step_mergeEdr=/) and $_='step_mergeEdr=1';
        (/retrOnOrbitOrSubOrbit=/) and $_='retrOnOrbitOrSubOrbit=1';
      };
      print $fhO "$_\n";
    }
    close($fhI) or die "$name\: Cannot close file $tmp.\nAbort";
    close($fhO) or die "$name\: Cannot close file $pcf.\nAbort";
    return $pcf;
  }

# ------------
  sub make_SCS
# ============
# $scs = make_SCS($tmpdir, $pcf, %options);
  {
    # $tmpdir is fully-qualified pathname of temporary working directory.
    # $pcf is fully-qualified PCF filepath for this run.
    # %options are the command line options, parsed and validated.
    # $scs is fully-qualified SCS filepath for this run.
    my ($tmpdir, $pcf) = (shift, shift);
    my (%options) = @_;
    my $name = (caller(0))[3];
    $name =~ s/main/$MAIN/g;
    my $setup = File::Spec -> catfile($HOME, 'setup');
    my $scripts = File::Spec -> catfile($HOME, 'scripts');
    foreach ($setup, $scripts) {
      -d or die "$name\:: MIRS subdirectory $_ not found.\nAbort";
    }
    my $tmp = File::Spec -> catfile($scripts, $options{sat}.'_scs_benchmark_hr.bash');
    -e $tmp or die "$MAIN\:: SCS template $tmp does not exist.\nAbort";
    my $scs = File::Spec -> catfile($tmpdir, 'cspp_'.$options{sat}.'_scs.bash');
    open(my $fhI, "<$tmp") or die "$name\: Cannot open $tmp for read.\nAbort";
    open(my $fhO, ">$scs") or die "$name\: Cannot open $scs for write.\nAbort";
    while (<$fhI>) { chomp;
      (/MIRS_ROOT=/)  and $_ = "MIRS_ROOT=$HOME";
      (/HDF4LIB=/)    and $_ = "HDF4LIB=$ENV{CSPP_MIRS_HOME}/lib";
      (/HDF5LIB=/)    and $_ = "HDF5LIB=$ENV{CSPP_MIRS_HOME}/lib";
      (/HDF5BIN=/)    and $_ = "HDF5BIN=$ENV{CSPP_MIRS_HOME}/bin";
      (/HDFEOSLIB=/)  and $_ = "HDFEOSLIB=$ENV{CSPP_MIRS_HOME}/lib";
      (/SZIPLIB=/)    and $_ = "SZIPLIB=$ENV{CSPP_MIRS_HOME}/lib";
      (/ZLIBLIB=/)    and $_ = "ZLIBLIB=/lib64";
      (/NETCDF4LIB=/) and $_ = "NETCDF4LIB=$ENV{CSPP_MIRS_HOME}/lib";
      (/benchmark_hr.bash/) and $_ ='. '.$pcf; # i.e. source the PCF
      (/..\/setup/) and s/..\/setup/$setup/g;
      (/..\/scripts/) and s/..\/scripts/$scripts/g;
      print $fhO "$_\n";
    }
    close($fhI) or die "$name\: Cannot close file $scs\nAbort";
    close($fhO) or die "$name\: Cannot close file $tmp\nAbort";
    chmod (0755, $scs) or die "$name\: Cannot chmod $scs.\nAbort";
    return $scs;
  }

# --------------------
  sub make_tmpdir_tree
# ====================
# $outdir = make_tmpdir_tree($sat);
  {
    # Make ControlData, InputsData and TestbedData directory trees under
    #   cwd - needed for processing $sat data with MIRS, and without
    #   clashing in the distribution directories (which will otherwise
    #   happen for two+ processes running concurrently).
    # $outdir is the directory that will contain output netCDF files.
    # Aborts within on fail.
    # It looks like the delivery from NOAA/STAR will make directories anyway,
    # but I want to be clear on the directories actually needed. The delivery
    # may make extra directories that appear to be unused.
    my $sat = shift;
    my $name = (caller(0))[3];
    $name =~ s/main/$MAIN/g;
    my @dirs = ();
    foreach (qw/ControlData InputsData ExternalData TestbedData/) {
      /ControlData/ and push(@dirs, File::Spec -> catfile($LATEST, $_));
      /InputsData/ and push(@dirs, File::Spec -> catfile($LATEST, $_));
      /ExternalData/ and do {
        foreach (qw/amsua atms/) {
          push(@dirs, File::Spec -> catfile($LATEST, 'ExternalData', 'gridNWP_sfr', $_));
      } };
      /TestbedData/ and do {
        my $d1 = $_;
        foreach (qw/DynamicData nedt Outputs/) {
          /DynamicData/ and do {
            my $d2 = $_;
            foreach (qw/fmsdr regress_retr sdr swe_climo tdr/) {
              my $d3 = $_;
              push(@dirs, File::Spec -> catfile($LATEST, $d1, $d2, $d3));
            } };
          /nedt/ and do {
            my $d2 = $_;
            my $d3 = undef;
            ($sat eq 'npp') and $d3 = 'npp_atms';
            ($sat eq 'n18') and $d3 = 'n18_amsua';
            ($sat eq 'n19') and $d3 = 'n19_amsua';
            ($sat eq 'n20') and $d3 = 'n20_atms';
            ($sat eq 'metopA') and $d3 = 'metopA_amsua';
            ($sat eq 'metopB') and $d3 = 'metopB_amsua';
            push(@dirs, File::Spec -> catfile($LATEST, $d1, $d2, $d3));
            ($sat eq 'npp') and next;
            ($sat eq 'n18') and $d3 = 'n18_mhs';
            ($sat eq 'n19') and $d3 = 'n19_mhs';
            ($sat eq 'n20') and next;
            ($sat eq 'metopA') and $d3 = 'metopA_mhs';
            ($sat eq 'metopB') and $d3 = 'metopB_mhs';
            push(@dirs, File::Spec -> catfile($LATEST, $d1, $d2, $d3));
            ($sat eq 'npp') and next;
            ($sat eq 'n18') and $d3 = 'n18_amsua_mhs';
            ($sat eq 'n19') and $d3 = 'n19_amsua_mhs';
            ($sat eq 'n20') and next;
            ($sat eq 'metopA') and $d3 = 'metopA_amsua_mhs';
            ($sat eq 'metopB') and $d3 = 'metopB_amsua_mhs';
            push(@dirs, File::Spec -> catfile($LATEST, $d1, $d2, $d3)); };
          /Outputs/ and do {
            my $d2 = $_;
            foreach (qw/dep edr nc/) {
              my $d3 = $_;
              my $d4 = undef;
              ($sat eq 'npp') and $d4 = 'npp_atms';
              ($sat eq 'n18') and $d4 = 'n18_amsua_mhs';
              ($sat eq 'n19') and $d4 = 'n19_amsua_mhs';
              ($sat eq 'n20') and $d4 = 'n20_atms';
              ($sat eq 'metopA') and $d4 = 'metopA_amsua_mhs';
              ($sat eq 'metopB') and $d4 = 'metopB_amsua_mhs';
              push(@dirs, File::Spec -> catfile($LATEST, $d1, $d2, $d3, $d4));
            } };
        } };
    }
    foreach (@dirs) { mkpath($_) or die "$name\: Cannot make dir $_\nAbort"; }
    return $dirs[$#dirs]; # last dir is netCDF output dir
  }

# --------------------------
  sub npp_granules_to_swaths
# ==========================
# %swaths = npp_granules_to_swaths(@files);
  {
    # Break list of files into groups that comprise contiguos overpass
    # swaths ready for the hdf5 aggregator (nagg) to concatenate.
    my @files = @_;
    my $name = (caller(0))[3];
    $name =~ s/main/$MAIN/g;
    my $time_window = 60 * $AGG_WINDOW_MINUTES;
    my %swaths = ();
    my ($YYYY, $MM, $DD, $hh, $mm, $found) = undef;
    foreach (@files) {
      $YYYY = substr($_, 11, 4);
      $MM = substr($_, 15, 2);
      $DD = substr($_, 17, 2);
      $hh = substr($_, 21, 2);
      $mm = substr($_, 23, 2);
      is_plausible_datetime($YYYY, $MM, $DD, $hh, $mm) or
        die "$name\: Date from npp file $_ implausible\nAbort";
      my $t = timegm_nocheck(0, $mm, $hh, $DD, $MM-1, $YYYY);
      $found = 0;
      foreach my $key (sort keys %swaths) {
        (abs($key - $t) < $time_window) and do {
          push(@{$swaths{$key}}, $_);
          $found = 1; last; };
      }
      ($found) or @{$swaths{$t}} = ($_);
    }
    return %swaths;
  }

# -----------------------
  sub populate_dynanc_gfs
# =======================
# populate_dynanc_gfs(\%dynanc_loc, \%options, @YYYYMMDDHH);
  {
    # %dynanc_loc is remote/local information for GFS dynamic ancillary data.
    # %options are the command line options.
    # @YYYYMMDDHH are the date/hour(s) we have established will be needed and
    # are already unique.
    # N.B. I am goin to 'warn' in here where I woould normally 'die' so that
    # if multiple days are being processed, there rmeains a chance the next
    # days will process.
    # During the execution of this function, the $PWD is assumed to be $tmpdir.
    # New problem - latency issue for near-real-time processing: The scripts
    # provided ask that I pick the last analysis time befoe the start of the
    # overpass, so if overpass starts at 1201Z, I will want to start with the
    # 12z analysis GFS files.  But if I am trying to process this overpass at,
    # say, 1500Z the 12z analysis GFS files will not yet be available (they
    # seem to show up about 4 hours after analysis time). This means that I
    # need to provide a fallback position to the previous 6 hour analysis time.
    # Plan is to fetch these by default (so that the local repo has them if
    # needed) and to "trick" the gfs_{atms,amsua}_sfr.bash scripts (provided
    # as part of MIRS DAP) by linking to 6-hour later forecasts of the 6-hour
    # earlier analysis.                                    JED    20-Jul-2018
    my %dynanc_loc = %{(shift)};
    my %options = %{(shift)};
    my @YYYYMMDDHH = @_;
    my $name = (caller(0))[3];
    $name =~ s/main/$MAIN/g;
    # Construct filenames to fetch
    my $preamble = "$dynanc_loc{'GFSPROTO'}$dynanc_loc{'GFSSITE'}";
    my $site_template = "$preamble/$dynanc_loc{'GFSREMOTE'}";
    my $file_template = "$dynanc_loc{'GFSTEMPLATE'}";
    my $local_template = "$dynanc_loc{'GFSLOCAL'}";
    my %tokens = ('%Y'=>4, '%y'=>2, '%m'=>2, '%j'=>3, '%d'=>2,
                  '%H'=>2, '%M'=>2, '%S'=>2, '%F'=>3, '%f'=>2);
    foreach my $YYYYMMDDHH (@YYYYMMDDHH) {
      my $YYYYJJJ = convert_YYYYMMDD_to_YYYYJJJ(substr($YYYYMMDDHH, 0, 8));
      my $check = "$YYYYMMDDHH$YYYYJJJ";
      my $proto = $file_template;
      my $local = $local_template;
      my $site = $site_template;
      foreach my $key (keys %tokens) {
        (grep {$_ eq $key} qw/%f %F/) and next;
        my $fmt = '%'.sprintf("0%sd", $tokens{$key});
        foreach my $mask (('%Y%m%d%HCCYY%j', 'CC%y%m%d%HCCYY%j')) {
          my $found = token_value($key, $check, $mask);
          ($found < 0) or do {
            my $value = sprintf($fmt, $found);
            $proto =~ s/$key/$value/g;
            $local =~ s/$key/$value/g;
            $site =~ s/$key/$value/g;
          };
        }
      }
      # This is slower that a wildcard fetch but required because of how
      # lftp_mirror() is coded to check specific files in $local before fetch.
      foreach my $ff ((0,3,6,9,12)) {
        my $file = $proto;
        my $fmt = '%'.sprintf("0%sd", $tokens{'%f'});
        my $value = sprintf($fmt, $ff);
        $file =~ s/%f/$value/g;
        #my $status = lftp_mirror($site, $file, $local, %options);
        # Get GFS files and make links just like I do for CSPP_NUCAPS
        my $waitawhile = $SLEEP_GFS_MINUTES;
        my $target = File::Spec -> catfile($local, $file);
        my $source = File::Spec -> catfile($GFSDATA, $file);
        # If I have a link already, assume all good, otherwise for a $file needed
        # for multiple overpasses (possible) I will get a lot of chatter from
        # lftp_mirror() repeatedly telling me "FOUND in local repo".
        -e $source or do {
          lftp_mirror($site, $file, $local, %options);
          -e $target and symlink($target, $source) or
            warn "$name\: WARNING: Cannot link to $target";
          ($options{quarantine}) and next; # don't need multiple attempts at local repo
          -e $source and next; # we have $target and link, so we are done for this $file
          #my $j = $waitawhile; # so $j is still in scope after the following loop
          for (my $j=$waitawhile; $j>0; $j--) {
            ($options{verbose} >= 1) and warn "$name\: SLEEPING at minute $j waiting for $file\n";
            sleep 60;
            lftp_mirror($site, $file, $local, %options);
            -e $target or next;
            symlink($target, $source) or warn "$name\: WARNING: Cannot link to $target";
            last;
          }
          #$waitawhile = $j;
        };
      }
      # Check what you have just linked to, because if you don't have them
      # all, you will need to repeat process for the 6-hour earlier analysis.
      my @linkname = ();
      my $count = 0;
      foreach my $ff ((0,3,6,9,12)) {
        my $file = $proto;
        my $fmt = '%'.sprintf("0%sd", $tokens{'%f'});
        my $value = sprintf($fmt, $ff);
        $file =~ s/%f/$value/g;
        my $source = File::Spec -> catfile($GFSDATA, $file);
        push(@linkname, $source);
        -e $source and $count++
      }
      ### print "\n\ncount:   $count\n\n\n";
      if ($count != 5) {
        # Get rid of the partially complete list of links to later avoid
        # failing to create links back to previous analysis time (but which
        # uses same link names to trick the MIRS GFS2BIN scripts).
        foreach (@linkname) { unlink($_); }
        # use $YYYYMMDDHH and $YYYYJJJ to figure out how to rollback 6 hours
        # you will need to compute linkname different to filename and make
        # the link even if linkname exists (i.e. count > 0 just not = 5).
        # If another overpass needs the same files, the loop above should
        # find the links and we don't get here a second time for the same 
        # reason.
        my $YYYYMMDD = substr($YYYYMMDDHH, 0, 8);
        my ($YYYY, $JJJ) = (substr($YYYYJJJ, 0, 4), substr($YYYYJJJ, 4, 3));
        my $HH = substr($YYYYMMDDHH, 8, 2);
        if ($HH eq '00') {
          $HH = 18;
          if ($JJJ eq '001') {
            $YYYY = sprintf("%04d", $YYYY - 1);
            $YYYYMMDD = $YYYY.'1231';
            $YYYYJJJ = convert_YYYYMMDD_to_YYYYJJJ($YYYYMMDD);
          }
        } else {
          $HH = sprintf("%02d", $HH - 6);
        }
        my $check = "$YYYYMMDD$HH$YYYYJJJ";
        my $proto = $file_template;
        my $local = $local_template;
        my $site = $site_template;
        foreach my $key (keys %tokens) {
          (grep {$_ eq $key} qw/%f %F/) and next;
          my $fmt = '%'.sprintf("0%sd", $tokens{$key});
          foreach my $mask (('%Y%m%d%HCCYY%j', 'CC%y%m%d%HCCYY%j')) {
            my $found = token_value($key, $check, $mask);
            ($found < 0) or do {
              my $value = sprintf($fmt, $found);
              $proto =~ s/$key/$value/g;
              $local =~ s/$key/$value/g;
              $site =~ s/$key/$value/g;
            };
          }
        }
        my $idx = 0;
        foreach my $ff ((6,9,12,15,18)) {
          my $file = $proto;
          my $fmt = '%'.sprintf("0%sd", $tokens{'%f'});
          my $value = sprintf($fmt, $ff);
          $file =~ s/%f/$value/g;
          # Get GFS files and make links just like I do for CSPP_NUCAPS
          my $waitawhile = $SLEEP_GFS_MINUTES;
          my $target = File::Spec -> catfile($local, $file);
          my $source = $linkname[$idx++];
          # If I have a link already, assume all good, otherwise for a $file needed
          # for multiple overpasses (possible) I will get a lot of chatter from
          # lftp_mirror() repeatedly telling me "FOUND in local repo".
          -e $source or do {
            lftp_mirror($site, $file, $local, %options);
            -e $target and symlink($target, $source) or
              warn "$name\: WARNING: Cannot link to $target";
            ($options{quarantine}) and next; # don't need multiple attempts at local repo
            -e $source and next; # we have $target and link, so we are done for this $file
            for (my $j=$waitawhile; $j>0; $j--) {
              ($options{verbose} >= 1) and warn "$name\: SLEEPING at minute $j waiting for $file\n";
              sleep 60;
              lftp_mirror($site, $file, $local, %options);
              -e $target or next;
              symlink($target, $source) or warn "$name\: WARNING: Cannot link to $target";
              last;
            }
          };
        }
      }
    }
    return;
  }

# ---------------
  sub pretty_dump
# ===============
# pretty_dump(%hash);
  {
    # Do a "pretty print" of %hash by effectively grepping Dumper().
    # Assumes standard module Data::Dumper has been loaded.
    my (%hash) = @_;
    foreach (split('\n', Dumper(\%hash))) {
      chomp;
      ((/VAR1/) or (/\}\;/)) and next;
      $_ =~ s/  //g; # shorten up the whitespace
      print "$_\n";
    }
  }

# ---------------
  sub token_value
# ===============
# $value = token_value($token, $file, $mask);  ...example follows...
# $value = token_value('%j', 't1.11332.1500.mod04.hdf', 't1.%y%j.%H%M.mod04.hdf');
  {
    # Takes the filename $file (within which tokens have already been expanded),
    # and the $mask (within which tokens are unexpanded) and, for a given $token
    # return the $value extracted from $file. Here is an example ...
    #   $value = token_value('%j', 't1.11332.1500.mod04.hdf', 't1.%y%j.%H%M.mod04.hdf');
    # which would return $value as '332'.
    # These are the supported tokens, returns the first occurance if more than one:
    #   %Y  Year with century. (1582-9999)
    #   %y  Year without century (00–99).
    #   %m  Month as decimal number (01–12).
    #   %j  Day of year as decimal number (001–366).
    #   %d  Day of the month as decimal number (01–31).
    #   %H  Hours as decimal number (00–23).
    #   %M  Minute as decimal number (00–59).
    #   %S  Second as decimal number (00–59).
    #   %F  3-digit forecast hour (000-999).
    #   %f  2-digit forecast hour (00-99).
    my $token = shift;
    my $file = shift;
    my $mask = shift;
    my $value = -1;
    my $expand = $mask;
    my @tokens = qw/  %Y %y %m  %j %d %H %S %M  %F %f/;
    my @widths =   (   4, 2, 2,  3, 2, 2, 2, 2,  3, 2);
    my $width = 0;
    for (my $j=0; $j<=$#tokens; $j++) {
      ($token eq $tokens[$j]) and do { $width = $widths[$j]; next; };
      my $blank = 'X' x $widths[$j];
      $expand =~ s/$tokens[$j]/$blank/g; 
    }
    my $i = index($expand, $token);
    if ($i > -1) {
      $value = substr($file, $i, $width);
    }
    return $value;
  }

# ---------------
  sub trim_string
# ===============
# $new_string = trim_string($old_string);
  {
    my $string = shift;
    $string =~ s/^\s+//;
    $string =~ s/\s+$//;
    return $string;
  }

# ---------
  sub usage
# =========
# usage();
  {
    my $name = (caller(0))[3];
    $name =~ s/main/$MAIN/g;
    open(my $fh, "<$0");
    while (<$fh>) { chomp;
      if ( /START-HELP/ ) {
        while (<$fh>) { chomp;
          ( /END-HELP/ ) and do { close($fh); exit(9); };
          my $help = (split(/\#/, $_))[1]; # remove leading hash
          $help and print "$help\n" or print "\n";
        }
      }
    }
    die "$name\: Unprecedented error!\nAbort";
  }

# --------------------
  sub validate_options
# ====================
# %options = validate_options($filemask, %options);
  {
    # $filemask is for globbing the nominated satellite data.
    # %options are the command line options to "validate".
    # Fatal errors in here terminate with die without showing usage() message.
    # The most important task is to convert $options{input_files} to a scalar
    # $options{input_dir} and filename list put back in $options{input_files}.
    my $filemask = shift;
    my (%options) = @_;
    my $name = (caller(0))[3];
    $name =~ s/main/$MAIN/g;
    my $globber = $options{input_files};
    (-d $options{input_files}) and do {
      my $base = $options{input_files};
      $globber = File::Spec -> catfile($base, $filemask); };
    my @input_files = sort(glob($globber));
    ($#input_files < 0) and die "$name\: INPUTFILES not found.\nAbort";
    $options{input_dir} = dirname(abs_path($input_files[0]));
    $options{input_files} = ();
    foreach (@input_files) { push(@{$options{input_files}}, basename($_)); }
    (-d $options{work}) or do {
      ($options{verbose} >=2) and print "Making WORK_DIR $options{work}\n";
      mkpath($options{work}) or
        die "$name\: WORK_DIR $options{work} cannot be made.\nAbort"; };
    (-w $options{work}) or
      die "$name\: WORK_DIR $options{work} not writable.\nAbort";
    (grep {$options{resolution} eq $_} qw/h H hi HI hI Hi/) and do {
      $options{resolution} = 'HI'; };
    (grep {$options{resolution} eq $_} qw/l L lo LO lO Lo/) and do {
      $options{resolution} = 'LO'; };
    (grep {$options{resolution} eq $_} qw/HI LO/) or
      die "$name\: RES $options{resolution} invalid resolution.\nAbort";
    (($options{sat} eq 'npp') and ($options{resolution} eq 'LO')) and
      die "$name\: RES $options{resolution} invalid for 'npp'.\nAbort";
    (($options{sat} eq 'n20') and ($options{resolution} eq 'LO')) and
      die "$name\: RES $options{resolution} invalid for 'n20'.\nAbort";
    (($options{block} >= 1) and ($options{block} <= 2)) or
      die "$name\: BLOCK $options{block} invalid ADL block.\nAbort";
    (($options{processors} < 1) or ($options{processors} > 32)) and
      die "$name\: PROCESSORS $options{processors} invalid number.\nAbort";
    return %options;
  }

# --------------------------
  sub verified_MIRS_filename
# ==========================
# my $veriname = verified_MIRS_filename($filename);
  {
    # $filename could be (for example) ...
    # NPR-MIRS-IMG_v11r3_n19_s201901072259000_e201901072310000_c201901141955130.nc
    # NPR-MIRS-IMG_v11r3_n19_s20190939000_e20190949000_c201901141957440.nc
    # NPR-MIRS-SND_v11r3_n19_s201901072259000_e201901072310000_c201901141955130.nc
    # NPR-MIRS-SND_v11r3_n19_s20190939000_e20190949000_c201901141957440.nc
    # ... but it is important that= the underscores are present as indicated !
    my $file = shift;
    my @parts = split(/\_/, basename($file));
    my ($Yf, $Yt) = first_and_last($file, 'ScanTime_year');
    my ($Mf, $Mt) = first_and_last($file, 'ScanTime_month');
    my ($Df, $Dt) = first_and_last($file, 'ScanTime_dom');
    my ($Uf, $Ut) = first_and_last($file, 'ScanTime_UTC');
    $parts[3] = 's'.datestring_for_filename($Yf, $Mf, $Df, $Uf);
    $parts[4] = 'e'.datestring_for_filename($Yt, $Mt, $Dt, $Ut);
    gmtime();
    $parts[5] = strftime("c%Y%m%d%H%M%S0.nc", $tm_sec, $tm_min, $tm_hour,
      $tm_mday, $tm_mon, $tm_year);
    my $fn = File::Spec -> catfile(dirname($file), join('_', @parts));
    return $fn;
  }
